from datetime import datetime, timedelta
from unittest import TestCase
from unittest.mock import patch

from celery.utils.time import timezone

from redbeat.schedules import rrule


@patch.object(rrule, 'now', datetime.utcnow)
@patch.object(rrule, 'utc_enabled', True)
@patch.object(rrule, 'tz', timezone.utc)
class test_rrule_remaining_estimate(TestCase):
    def test_freq(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow() + timedelta(minutes=1))
        eta_from_now = r.remaining_estimate(datetime.utcnow())
        eta_after_one_minute = r.remaining_estimate(datetime.utcnow() + timedelta(minutes=1))
        self.assertTrue(eta_from_now.total_seconds() > 0)
        self.assertTrue(eta_after_one_minute.total_seconds() > 0)

    def test_freq__with_single_count(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow() + timedelta(minutes=1), count=1)
        eta_from_now = r.remaining_estimate(datetime.utcnow())
        eta_after_one_minute = r.remaining_estimate(datetime.utcnow() + timedelta(minutes=1))
        self.assertTrue(eta_from_now.total_seconds() > 0)
        self.assertEqual(eta_after_one_minute, None)

    def test_freq__with_multiple_count(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow() + timedelta(minutes=1), count=2)
        eta_from_now = r.remaining_estimate(datetime.utcnow())
        eta_after_one_minute = r.remaining_estimate(datetime.utcnow() + timedelta(minutes=1))
        eta_after_two_minutes = r.remaining_estimate(datetime.utcnow() + timedelta(minutes=2))
        self.assertTrue(eta_from_now.total_seconds() > 0)
        self.assertTrue(eta_after_one_minute.total_seconds() > 0)
        self.assertEqual(eta_after_two_minutes, None)


@patch.object(rrule, 'now', datetime.utcnow)
@patch.object(rrule, 'utc_enabled', True)
@patch.object(rrule, 'tz', timezone.utc)
class test_rrule_is_due(TestCase):
    def test_freq__starting_now(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow())
        # Assuming job was never run, i.e. last_run_at == epoch
        is_due, next = r.is_due(datetime(1970, 1, 1))
        self.assertTrue(is_due)
        self.assertTrue(next > 0)

    def test_freq__starts_after_one_minute(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow() + timedelta(minutes=1))
        is_due, next = r.is_due(datetime.utcnow())
        self.assertFalse(is_due)
        self.assertTrue(next > 0)

    def test_freq__with_single_count(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow(), count=1)
        # If job was never run, it should be due and have
        # no ETA for the following occurrence.
        is_due, next = r.is_due(datetime(1970, 1, 1))
        self.assertTrue(is_due)
        self.assertEqual(next, None)
        # It should not be due if it was already run once.
        is_due, next = r.is_due(datetime.utcnow())
        self.assertFalse(is_due)
        self.assertEqual(next, None)

    def test_freq__with_multiple_count(self):
        r = rrule('MINUTELY', dtstart=datetime.utcnow(), count=2)
        is_due, next = r.is_due(datetime(1970, 1, 1))
        self.assertTrue(is_due)
        self.assertTrue(next > 0)
        # There should still be one more occurrence remaining
        # if it was run once after dtstart.
        is_due, next = r.is_due(datetime.utcnow())
        self.assertFalse(is_due)
        self.assertTrue(next > 0)
        # There should be no more occurrences after one minute.
        is_due, next = r.is_due(datetime.utcnow() + timedelta(minutes=1))
        self.assertFalse(is_due)
        self.assertEqual(next, None)



from unittest import mock

from redbeat.schedulers import RedBeatConfig
from tests.basecase import AppCase


class test_RedBeatConfig(AppCase):
    def setup(self):
        self.conf = RedBeatConfig(self.app)

    def test_app(self):
        self.assertEqual(self.app, self.conf.app)

    def test_lock_timeout(self):
        # the config only has the lock_timeout if it was overidden
        # via the REDBEAT_LOCK_TIMEOUT, see test_scheduler.py for real test
        self.assertEqual(self.conf.lock_timeout, None)

    def test_key_prefix_default(self):
        self.assertEqual(self.conf.key_prefix, 'redbeat:')

    def test_other_keys(self):
        self.assertEqual(self.conf.schedule_key, self.conf.key_prefix + ':schedule')
        self.assertEqual(self.conf.statics_key, self.conf.key_prefix + ':statics')
        self.assertEqual(self.conf.lock_key, self.conf.key_prefix + ':lock')

    def test_key_prefix_override(self):
        self.app.conf.redbeat_key_prefix = 'test-prefix:'
        self.conf = RedBeatConfig(self.app)
        self.assertEqual(self.conf.key_prefix, 'test-prefix:')

    def test_lock_key_might_be_set_to_none(self):
        self.app.conf.redbeat_lock_key = None
        self.conf = RedBeatConfig(self.app)
        self.assertEqual(self.conf.lock_key, None)

    def test_lock_key_override(self):
        self.app.conf.redbeat_lock_key = ":custom"
        self.conf = RedBeatConfig(self.app)
        self.assertEqual(self.conf.lock_key, 'redbeat::custom')

    def test_schedule(self):
        schedule = {'foo': 'bar'}
        self.conf.schedule = schedule
        self.assertEqual(self.conf.schedule, schedule)

    @mock.patch('warnings.warn')
    def test_either_or(self, warn_mock):
        broker_url = self.conf.either_or('BROKER_URL')
        self.assertTrue(warn_mock.called)
        self.assertEqual(broker_url, self.app.conf.broker_url)



from celery.utils.time import maybe_make_aware

from redbeat.decoder import from_timestamp, to_timestamp
from tests.basecase import RedBeatCase


class Test_utils(RedBeatCase):
    def test_roundtrip(self):
        now = self.app.now()
        # 3.x returns naive, but 4.x returns aware
        now = maybe_make_aware(now)

        roundtripped = from_timestamp(to_timestamp(now))

        # we lose microseconds in the roundtrip, so we need to ignore them
        now = now.replace(microsecond=0)

        self.assertEqual(now, roundtripped)



import calendar
import json
from datetime import datetime, timedelta, timezone

from celery.schedules import crontab, schedule
from dateutil.rrule import weekday

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from .schedules import rrule


def to_timestamp(dt):
    """convert local tz aware datetime to seconds since the epoch"""
    return calendar.timegm(dt.utctimetuple())


def get_utcoffset_minutes(dt):
    """calculates timezone utc offset, returns minutes relative to utc"""
    utcoffset = dt.utcoffset()

    # Python 3: utcoffset / timedelta(minutes=1)
    return utcoffset.total_seconds() / 60 if utcoffset else 0


def from_timestamp(seconds, tz_minutes=0):
    """convert seconds since the epoch to an UTC aware datetime"""
    tz = timezone(timedelta(minutes=tz_minutes)) if tz_minutes else timezone.utc
    return datetime.fromtimestamp(seconds, tz=tz)


class RedBeatJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kargs):
        super().__init__(object_hook=self.dict_to_object, *args, **kargs)

    def dict_to_object(self, d):
        if '__type__' not in d:
            return d

        objtype = d.pop('__type__')

        if objtype == 'datetime':
            zone = d.pop('timezone', 'UTC')
            try:
                tzinfo = timezone(timedelta(minutes=(zone / 60)))
            except TypeError:
                tzinfo = zoneinfo.ZoneInfo(zone)
            return datetime(tzinfo=tzinfo, **d)

        if objtype == 'interval':
            return schedule(run_every=d['every'], relative=d['relative'])

        if objtype == 'crontab':
            return crontab(**d)

        if objtype == 'weekday':
            return weekday(**d)

        if objtype == 'rrule':
            # Decode timestamp values into datetime objects
            for key, tz_key in [('dtstart', 'dtstart_tz'), ('until', 'until_tz')]:
                timestamp = d.get(key)
                tz_minutes = d.pop(tz_key, 0)
                if timestamp is not None:
                    d[key] = from_timestamp(timestamp, tz_minutes)
            return rrule(**d)

        d['__type__'] = objtype

        return d


class RedBeatJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            if obj.tzinfo is None:
                timezone = 'UTC'
            elif hasattr(obj.tzinfo, 'key'):
                timezone = obj.tzinfo.key
            else:
                timezone = obj.utcoffset().total_seconds()

            return {
                '__type__': 'datetime',
                'year': obj.year,
                'month': obj.month,
                'day': obj.day,
                'hour': obj.hour,
                'minute': obj.minute,
                'second': obj.second,
                'microsecond': obj.microsecond,
                'timezone': timezone,
            }

        if isinstance(obj, crontab):
            return {
                '__type__': 'crontab',
                'minute': obj._orig_minute,
                'hour': obj._orig_hour,
                'day_of_week': obj._orig_day_of_week,
                'day_of_month': obj._orig_day_of_month,
                'month_of_year': obj._orig_month_of_year,
            }

        if isinstance(obj, rrule):
            res = {
                '__type__': 'rrule',
                'freq': obj.freq,
                'interval': obj.interval,
                'wkst': obj.wkst,
                'count': obj.count,
                'bysetpos': obj.bysetpos,
                'bymonth': obj.bymonth,
                'bymonthday': obj.bymonthday,
                'byyearday': obj.byyearday,
                'byeaster': obj.byeaster,
                'byweekno': obj.byweekno,
                'byweekday': obj.byweekday,
                'byhour': obj.byhour,
                'byminute': obj.byminute,
                'bysecond': obj.bysecond,
            }

            # Convert datetime objects to timestamps
            if obj.dtstart:
                res['dtstart'] = to_timestamp(obj.dtstart)
                res['dtstart_tz'] = get_utcoffset_minutes(obj.dtstart)

            if obj.until:
                res['until'] = to_timestamp(obj.until)
                res['until_tz'] = get_utcoffset_minutes(obj.until)

            return res

        if isinstance(obj, weekday):
            return {'__type__': 'weekday', 'wkday': obj.weekday}

        if isinstance(obj, schedule):
            return {
                '__type__': 'interval',
                'every': obj.run_every.total_seconds(),
                'relative': bool(obj.relative),
            }

        return super().default(obj)



import ssl
import unittest
from copy import deepcopy
from datetime import datetime, timedelta
from unittest import mock
from unittest.mock import MagicMock, Mock, patch

import pytz
from celery.beat import DEFAULT_MAX_INTERVAL
from celery.schedules import schedstate, schedule
from celery.utils.time import maybe_timedelta
from redis.exceptions import ConnectionError

from redbeat import RedBeatScheduler
from redbeat.schedulers import RedBeatSchedulerEntry, acquire_distributed_beat_lock, get_redis
from tests.basecase import AppCase, RedBeatCase


class mocked_schedule(schedule):
    def __init__(self, remaining):
        self._remaining = maybe_timedelta(remaining)
        self.run_every = timedelta(seconds=1)
        self.nowfun = datetime.utcnow

    def remaining_estimate(self, last_run_at):
        return self._remaining


class mocked_expired_schedule(schedule):
    def __init__(self):
        self.nowfun = datetime.utcnow
        self.run_every = Mock()
        self.run_every.total_seconds.return_value = 1

    def remaining_estimate(self, last_run_at):
        return None

    def is_due(self, last_run_at):
        return schedstate(False, None)


due_now = mocked_schedule(0)
due_next = mocked_schedule(1)


class RedBeatSchedulerTestBase(RedBeatCase):
    def setUp(self):
        super().setUp()
        self.s = RedBeatScheduler(app=self.app)
        self.due_later = mocked_schedule(self.s.max_interval * 10)
        self.send_task = patch.object(self.s, 'send_task')
        self.send_task.start()

    def tearDown(self):
        self.send_task.stop()


class test_RedBeatScheduler_schedule(RedBeatSchedulerTestBase):
    def test_empty_schedule(self):
        self.assertEqual(self.s.schedule, {})

    def test_schedule_includes_current_and_next(self):
        due = self.create_entry(name='due', s=due_now).save()
        up_next = self.create_entry(name='up_next', s=due_next).save()
        up_next2 = self.create_entry(name='up_next2', s=due_next).save()
        later = self.create_entry(name='later', s=self.due_later).save()

        schedule = self.s.schedule

        self.assertEqual(len(schedule), 2)

        self.assertIn(due.name, schedule)
        self.assertEqual(due.key, schedule[due.name].key)

        self.assertIn(up_next.name, schedule)
        self.assertEqual(up_next.key, schedule[up_next.name].key)

        self.assertNotIn(up_next2.name, schedule)
        self.assertNotIn(later.name, schedule)


class test_RedBeatScheduler_tick(RedBeatSchedulerTestBase):
    def setUp(self):
        super().setUp()
        self.s.lock_key = None

    def test_empty(self):
        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            self.assertFalse(send_task.called)

        self.assertEqual(sleep, self.s.max_interval)

    def test_due_next_never_run(self):
        self.create_entry(name='next', s=due_next).save()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            self.assertFalse(send_task.called)

        self.assertLess(sleep, 1.0)

    def test_due_next_just_ran(self):
        e = self.create_entry(name='next', s=due_next)
        e.save().reschedule()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            self.assertFalse(send_task.called)
        self.assertLess(0.8, sleep)
        self.assertLess(sleep, 1.0)

    @unittest.skip('test is breaking and unsure what the correct behaviour should be')
    def test_due_next_never_run_tz_positive(self):
        self.app.timezone = pytz.timezone('Europe/Moscow')

        e = self.create_entry(name='next', s=due_next).save()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            send_task.assert_called_with(e.task, e.args, e.kwargs, **self.s._maybe_due_kwargs)
            # would be more correct to
            # self.assertFalse(send_task.called)

        self.assertEqual(sleep, 1.0)
        self.app.timezone = pytz.utc

    @unittest.skip('test is breaking and unsure what the correct behaviour should be')
    def test_due_next_never_run_tz_negative(self):
        self.app.timezone = pytz.timezone('America/Chicago')

        e = self.create_entry(name='next', s=due_next).save()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            send_task.assert_called_with(e.task, e.args, e.kwargs, **self.s._maybe_due_kwargs)
            # would be more correct to
            # self.assertFalse(send_task.called)

        self.assertEqual(sleep, 1.0)
        self.app.timezone = pytz.utc

    def test_due_later_task_never_run(self):
        self.create_entry(s=self.due_later).save()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            self.assertFalse(send_task.called)

        self.assertEqual(sleep, self.s.max_interval)

    def test_due_now(self):
        e = self.create_entry(
            name='now',
            args=[],
            s=due_now,
            last_run_at=datetime.utcnow() - timedelta(seconds=1),
        ).save()

        with patch.object(self.s, 'send_task') as send_task:
            sleep = self.s.tick()
            send_task.assert_called_with(e.task, e.args, e.kwargs, **self.s._maybe_due_kwargs)

        self.assertEqual(sleep, 1.0)

    def test_old_static_entries_are_removed(self):
        redis = self.app.redbeat_redis
        schedule = {'test': {'task': 'test', 'schedule': mocked_schedule(42)}}
        self.app.redbeat_conf.schedule = schedule
        self.s.setup_schedule()

        self.assertIn('test', self.s.schedule)
        self.assertIn('test', redis.smembers(self.app.redbeat_conf.statics_key))

        self.app.redbeat_conf.schedule = {}
        self.s.setup_schedule()

        self.assertNotIn('test', self.s.schedule)
        self.assertNotIn('test', redis.smembers(self.app.redbeat_conf.statics_key))

    def test_lock_timeout(self):
        self.assertEqual(self.s.lock_timeout, self.s.max_interval * 5)

    def test_lock_acquisition_failed_during_startup(self):
        self.s.lock_key = 'lock-key'
        self.s.lock = None
        with self.assertRaises(AttributeError):
            self.s.tick()


class TestRedBeatSchedulerUpdateFromDict(RedBeatSchedulerTestBase):
    @patch.object(RedBeatSchedulerEntry, "from_key")
    def test_update_from_dict_fetch_redis_entry(self, mock_from_key: MagicMock) -> None:
        mock_entry_from_redis_key = mock_from_key.return_value

        self.s.update_from_dict(
            dict_={'task_name': {'task': 'tasks.task_name', 'schedule': timedelta(seconds=30)}}
        )

        mock_from_key.assert_called_once_with("rb-tests:task_name", app=self.app)

        mock_entry_from_redis_key.save.assert_called_once()

    @patch.object(RedBeatSchedulerEntry, "from_key")
    @patch.object(RedBeatSchedulerEntry, "save")
    def test_update_from_dict(self, mock_entry_save: MagicMock, mock_from_key: MagicMock) -> None:
        mock_from_key.side_effect = [KeyError()]

        self.s.update_from_dict(
            dict_={'task_name': {'task': 'tasks.task_name', 'schedule': timedelta(seconds=30)}}
        )

        mock_entry_save.assert_called_once()


class NotSentinelRedBeatCase(AppCase):
    config_dict = {
        'BROKER_URL': 'redis://',
    }

    def setup(self):
        self.app.conf.update(self.config_dict)

    def test_sentinel_scheduler(self):
        redis_client = get_redis(app=self.app)
        assert 'Sentinel' not in str(redis_client.connection_pool)


class ClusterRedBeatCase(AppCase):
    config_dict = {
        'BROKER_URL': 'redis://',
        'REDBEAT_REDIS_OPTIONS': {
            'cluster': True,
        },
    }

    def setup(self):
        self.app.conf.update(self.config_dict)

    def test_sentinel_scheduler(self):
        # Fake redis doesn't really support redis cluster, but let's just make sure it was used.
        with mock.patch('redis.RedisCluster.from_url') as from_url:
            get_redis(app=self.app)
            self.assertTrue(from_url.called)


class SentinelRedBeatCase(AppCase):
    config_dict = {
        'REDBEAT_KEY_PREFIX': 'rb-tests:',
        'redbeat_key_prefix': 'rb-tests:',
        'BROKER_URL': 'redis-sentinel://redis-sentinel:26379/0',
        'CELERY_RESULT_BACKEND': 'redis-sentinel://redis-sentinel:26379/1',
    }
    BROKER_TRANSPORT_OPTIONS = {
        'sentinels': [('192.168.1.1', 26379), ('192.168.1.2', 26379), ('192.168.1.3', 26379)],
        'service_name': 'master',
        'socket_timeout': 0.1,
    }

    def setup(self):  # celery3
        self.app.conf.add_defaults(deepcopy(self.config_dict))

    def test_sentinel_scheduler(self):
        self.app.conf.update({'BROKER_TRANSPORT_OPTIONS': self.BROKER_TRANSPORT_OPTIONS})
        redis_client = get_redis(app=self.app)
        assert 'Sentinel' in str(redis_client.connection_pool)

    def test_sentinel_scheduler_options(self):
        for options in [
            "BROKER_TRANSPORT_OPTIONS",
            "redbeat_redis_options",
            "REDBEAT_REDIS_OPTIONS",
        ]:
            config = deepcopy(self.config_dict)
            config[options] = self.BROKER_TRANSPORT_OPTIONS
            self.app.conf.clear()
            self.app.conf.update(config)
            redis_client = get_redis(app=self.app)
            assert 'Sentinel' in str(redis_client.connection_pool)


class SeparateOptionsForSchedulerCase(AppCase):
    config_dict = {
        'REDBEAT_KEY_PREFIX': 'rb-tests:',
        'REDBEAT_REDIS_URL': 'redis-sentinel://redis-sentinel:26379/0',
        'REDBEAT_REDIS_OPTIONS': {
            'sentinels': [('192.168.1.1', 26379), ('192.168.1.2', 26379), ('192.168.1.3', 26379)],
            'password': '123',
            'service_name': 'master',
            'socket_timeout': 0.1,
        },
        'CELERY_RESULT_BACKEND': 'redis-sentinel://redis-sentinel:26379/1',
    }

    def setup(self):  # celery3
        self.app.conf.add_defaults(deepcopy(self.config_dict))

    def test_sentinel_scheduler(self):
        redis_client = get_redis(app=self.app)
        assert 'Sentinel' in str(redis_client.connection_pool)


class SSLConnectionToRedis(AppCase):
    config_dict = {
        'REDBEAT_KEY_PREFIX': 'rb-tests:',
        'REDBEAT_REDIS_URL': 'rediss://redishost:26379/0',
        'REDBEAT_REDIS_OPTIONS': {
            'password': '123',
            'service_name': 'master',
            'socket_timeout': 0.1,
        },
        'CELERY_RESULT_BACKEND': 'redis-sentinel://redis-sentinel:26379/1',
        'REDBEAT_REDIS_USE_SSL': {
            'ssl_cert_reqs': ssl.CERT_REQUIRED,
            'ssl_keyfile': '/path/to/file.key',
            'ssl_certfile': '/path/to/cert.pem',
            'ssl_ca_certs': '/path/to/ca.pem',
        },
    }

    def setup(self):  # celery3
        self.app.conf.add_defaults(deepcopy(self.config_dict))

    def test_ssl_connection_scheduler(self):
        redis_client = get_redis(app=self.app)
        assert 'SSLConnection' in str(redis_client.connection_pool)
        assert redis_client.connection_pool.connection_kwargs['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert redis_client.connection_pool.connection_kwargs['ssl_keyfile'] == '/path/to/file.key'
        assert redis_client.connection_pool.connection_kwargs['ssl_certfile'] == '/path/to/cert.pem'
        assert redis_client.connection_pool.connection_kwargs['ssl_ca_certs'] == '/path/to/ca.pem'


class SSLConnectionToRedisDefaultBrokerSSL(AppCase):
    config_dict = {
        'REDBEAT_KEY_PREFIX': 'rb-tests:',
        'REDBEAT_REDIS_URL': 'rediss://redishost:26379/0',
        'REDBEAT_REDIS_OPTIONS': {
            'password': '123',
            'service_name': 'master',
            'socket_timeout': 0.1,
        },
        'CELERY_RESULT_BACKEND': 'redis-sentinel://redis-sentinel:26379/1',
        'BROKER_USE_SSL': {
            'ssl_cert_reqs': ssl.CERT_REQUIRED,
            'ssl_keyfile': '/path/to/file.key',
            'ssl_certfile': '/path/to/cert.pem',
            'ssl_ca_certs': '/path/to/ca.pem',
        },
    }

    def setup(self):  # celery3
        self.app.conf.add_defaults(deepcopy(self.config_dict))

    def test_ssl_connection_scheduler(self):
        redis_client = get_redis(app=self.app)
        assert 'SSLConnection' in str(redis_client.connection_pool)
        assert redis_client.connection_pool.connection_kwargs['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert redis_client.connection_pool.connection_kwargs['ssl_keyfile'] == '/path/to/file.key'
        assert redis_client.connection_pool.connection_kwargs['ssl_certfile'] == '/path/to/cert.pem'
        assert redis_client.connection_pool.connection_kwargs['ssl_ca_certs'] == '/path/to/ca.pem'


class SSLConnectionToRedisNoCerts(AppCase):
    config_dict = {
        'REDBEAT_KEY_PREFIX': 'rb-tests:',
        'REDBEAT_REDIS_URL': 'rediss://redishost:26379/0',
        'REDBEAT_REDIS_OPTIONS': {
            'password': '123',
            'service_name': 'master',
            'socket_timeout': 0.1,
        },
        'CELERY_RESULT_BACKEND': 'redis-sentinel://redis-sentinel:26379/1',
        'REDBEAT_REDIS_USE_SSL': True,
    }

    def setup(self):  # celery3
        self.app.conf.add_defaults(deepcopy(self.config_dict))

    def test_ssl_connection_scheduler(self):
        redis_client = get_redis(app=self.app)
        assert 'SSLConnection' in str(redis_client.connection_pool)
        assert redis_client.connection_pool.connection_kwargs['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert 'ssl_keyfile' not in redis_client.connection_pool.connection_kwargs
        assert 'ssl_certfile' not in redis_client.connection_pool.connection_kwargs
        assert 'ssl_ca_certs' not in redis_client.connection_pool.connection_kwargs


class RedBeatLockTimeoutDefaultValues(RedBeatCase):
    def test_no_values(self):
        scheduler = RedBeatScheduler(app=self.app)
        assert DEFAULT_MAX_INTERVAL * 5 == scheduler.lock_timeout
        assert DEFAULT_MAX_INTERVAL == scheduler.max_interval


class RedBeatLockTimeoutCustomMaxInterval(RedBeatCase):
    config_dict = {
        'beat_max_loop_interval': 5,
    }

    def test_no_lock_timeout(self):
        scheduler = RedBeatScheduler(app=self.app)
        assert self.config_dict['beat_max_loop_interval'] * 5 == scheduler.lock_timeout
        assert self.config_dict['beat_max_loop_interval'] == scheduler.max_interval


class RedBeatLockTimeoutCustomAll(RedBeatCase):
    config_dict = {
        'beat_max_loop_interval': 7,
        'redbeat_lock_timeout': 13,
    }

    def test_custom_lock_timeout(self):
        scheduler = RedBeatScheduler(app=self.app)
        assert self.config_dict['beat_max_loop_interval'] == scheduler.max_interval
        assert self.config_dict['redbeat_lock_timeout'] == scheduler.lock_timeout


class RedBeatStartupAcquiresLock(RedBeatSchedulerTestBase):
    def setUp(self):
        super().setUp()
        self.sender = Mock(scheduler=self.s)

    def test_acquires_lock(self):
        acquire_distributed_beat_lock(self.sender)
        self.assertTrue(self.s.lock.owned())

    def test_connection_error(self):
        self.redis_server.connected = False
        with self.assertRaises(ConnectionError):
            acquire_distributed_beat_lock(self.sender)



import json
from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch

from celery.utils.time import maybe_make_aware

from redbeat import RedBeatSchedulerEntry
from redbeat.decoder import RedBeatJSONDecoder, from_timestamp, to_timestamp
from tests.basecase import RedBeatCase

CELERY_CONFIG_DEFAULT_KWARGS = {}


class test_RedBeatEntry(RedBeatCase):
    def test_basic_save(self):
        e = self.create_entry()
        e.save()

        expected = {
            'name': 'test',
            'task': 'tasks.test',
            'schedule': e.schedule,
            'args': None,
            'kwargs': CELERY_CONFIG_DEFAULT_KWARGS,
            'options': {},
            'enabled': True,
        }
        expected_key = self.app.redbeat_conf.key_prefix + 'test'

        redis = self.app.redbeat_redis
        value = redis.hget(expected_key, 'definition')
        self.assertEqual(expected, json.loads(value, cls=RedBeatJSONDecoder))
        self.assertEqual(redis.zrank(self.app.redbeat_conf.schedule_key, e.key), 0)
        self.assertEqual(redis.zscore(self.app.redbeat_conf.schedule_key, e.key), e.score)

    def test_from_key_nonexistent_key(self):
        with self.assertRaises(KeyError):
            RedBeatSchedulerEntry.from_key('doesntexist', self.app)

    def test_from_key_missing_meta(self):
        initial = self.create_entry().save()

        loaded = RedBeatSchedulerEntry.from_key(initial.key, self.app)
        self.assertEqual(initial.task, loaded.task)
        self.assertIsNotNone(loaded.last_run_at)

    def test_next(self):
        initial = self.create_entry().save()
        now = self.app.now()
        # 3.x is naive but 4.x is aware
        now = maybe_make_aware(now)

        n = initial.next(last_run_at=now)

        self.assertIsNotNone(now.tzinfo)
        self.assertEqual(n.last_run_at, now)
        self.assertEqual(initial.total_run_count + 1, n.total_run_count)

        # updated meta was stored into redis
        loaded = RedBeatSchedulerEntry.from_key(initial.key, app=self.app)
        self.assertEqual(loaded.last_run_at, now)
        self.assertEqual(loaded.total_run_count, initial.total_run_count + 1)

        # new entry updated the schedule
        redis = self.app.redbeat_redis
        self.assertEqual(redis.zscore(self.app.redbeat_conf.schedule_key, n.key), n.score)

    def test_next_only_update_last_run_at(self):
        initial = self.create_entry()

        n = initial.next(only_update_last_run_at=True)
        self.assertGreater(n.last_run_at, initial.last_run_at)
        self.assertEqual(n.total_run_count, initial.total_run_count)

    def test_delete(self):
        initial = self.create_entry()
        initial.save()

        e = RedBeatSchedulerEntry.from_key(initial.key, app=self.app)
        e.delete()

        exists = self.app.redbeat_redis.exists(initial.key)
        self.assertFalse(exists)

        score = self.app.redbeat_redis.zrank(self.app.redbeat_conf.schedule_key, initial.key)
        self.assertIsNone(score)

    def test_due_at_never_run(self):
        entry = self.create_entry(last_run_at=datetime.min)

        before = entry._default_now()
        due_at = entry.due_at
        after = entry._default_now()

        self.assertLess(before, due_at)
        self.assertLess(due_at, after)

    def test_due_at(self):
        entry = self.create_entry()

        now = entry._default_now()

        entry.last_run_at = now
        due_at = entry.due_at

        self.assertLess(now, due_at)
        self.assertLess(due_at, now + entry.schedule.run_every)

    def test_due_at_overdue(self):
        last_run_at = self.app.now() - timedelta(hours=10)
        entry = self.create_entry(last_run_at=last_run_at)

        before = entry._default_now()
        due_at = entry.due_at

        self.assertLess(last_run_at, due_at)
        self.assertGreater(due_at, before)

    def test_score(self):
        run_every = 61 * 60
        entry = self.create_entry(run_every=run_every)
        entry = entry._next_instance()

        score = entry.score
        expected = entry.last_run_at + timedelta(seconds=run_every)
        expected = expected.replace(microsecond=0)  # discard microseconds, lost in timestamp
        # 3.x returns naive, but 4.x returns aware
        expected = maybe_make_aware(expected)

        self.assertEqual(score, to_timestamp(expected))
        self.assertEqual(expected, from_timestamp(score))

    def test_generate_key(self) -> None:
        entry = self.create_entry()

        key = entry.generate_key(app=self.app, name="mock_task_name")

        self.assertEqual(key, "redbeat:mock_task_name")

    @patch.object(RedBeatSchedulerEntry, "generate_key")
    def test_key(self, mock_generate_key: MagicMock) -> None:
        entry = self.create_entry()

        key = entry.key

        mock_generate_key.assert_called_once_with(app=self.app, name="test")

        self.assertEqual(key, mock_generate_key.return_value)



import json
from datetime import datetime, timedelta, timezone
from unittest import TestCase

from celery.schedules import crontab, schedule
from dateutil import rrule as dateutil_rrule

from redbeat.decoder import RedBeatJSONDecoder, RedBeatJSONEncoder
from redbeat.schedules import rrule

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo


class JSONTestCase(TestCase):
    def dumps(self, d):
        return json.dumps(d, cls=RedBeatJSONEncoder)

    def loads(self, d):
        return json.loads(d, cls=RedBeatJSONDecoder)

    def datetime_as_dict(self, **kwargs):
        d = {
            'year': 2015,
            'month': 12,
            'day': 30,
            'hour': 12,
            'minute': 59,
            'second': 22,
            'microsecond': 333,
        }
        d.update(kwargs)
        return d

    def now(self, **kwargs):
        d = self.datetime_as_dict(**kwargs)
        dt = datetime(d.pop('year'), d.pop('month'), d.pop('day'), **d)
        return dt

    def schedule(self, **kwargs):
        d = {
            '__type__': 'interval',
            'every': 60.0,
            'relative': False,
        }
        d.update(kwargs)
        return d

    def crontab(self, **kwargs):
        d = {
            '__type__': 'crontab',
            'minute': '*',
            'hour': '*',
            'day_of_week': '*',
            'day_of_month': '*',
            'month_of_year': '*',
        }
        d.update(kwargs)
        return d

    def rrule(self, **kwargs):
        d = {
            '__type__': 'rrule',
            'freq': 5,
            'dtstart': 1451480362,
            'dtstart_tz': 0,
            'interval': 1,
            'wkst': None,
            'count': 1,
            'until': 1451566762,
            'until_tz': 0,
            'bysetpos': None,
            'bymonth': None,
            'bymonthday': None,
            'byyearday': None,
            'byeaster': None,
            'byweekno': None,
            'byweekday': None,
            'byhour': None,
            'byminute': None,
            'bysecond': None,
        }
        d.update(kwargs)
        return d

    def weekday(self, **kwargs):
        d = {
            '__type__': 'weekday',
            'wkday': 5,
        }
        d.update(kwargs)
        return d


class RedBeatJSONEncoderTestCase(JSONTestCase):
    def test_datetime_no_tz(self):
        dt = self.now()
        result = self.dumps(dt)

        expected = self.datetime_as_dict(__type__='datetime', timezone='UTC')
        self.assertEqual(json.loads(result), expected)

    def test_datetime_with_tz(self):
        dt = self.now(tzinfo=zoneinfo.ZoneInfo('Asia/Shanghai'))
        result = self.dumps(dt)

        expected = self.datetime_as_dict(timezone='Asia/Shanghai', __type__='datetime')
        self.assertEqual(json.loads(result), expected)

    def test_datetime_with_fixedoffset(self):
        dt = self.now(tzinfo=timezone(timedelta(hours=4)))
        result = self.dumps(dt)

        expected = self.datetime_as_dict(timezone=4 * 60 * 60.0)
        expected['__type__'] = 'datetime'
        self.assertEqual(json.loads(result), expected)

    def test_schedule(self):
        s = schedule(run_every=60.0)
        result = self.dumps(s)
        self.assertEqual(result, json.dumps(self.schedule(every=60.0)))

    def test_crontab(self):
        c = crontab()
        result = self.dumps(c)
        self.assertEqual(result, json.dumps(self.crontab()))

    def test_rrule(self):
        r = rrule(
            'MINUTELY',
            dtstart=datetime(2015, 12, 30, 12, 59, 22, tzinfo=timezone.utc),
            until=datetime(2015, 12, 31, 12, 59, 22, tzinfo=timezone.utc),
            count=1,
        )
        result = self.dumps(r)
        self.assertEqual(json.loads(result), self.rrule())

    def test_rrule_timezone(self):
        tz = zoneinfo.ZoneInfo('US/Eastern')

        start1 = datetime(2015, 12, 30, 12, 59, 22, tzinfo=timezone.utc)
        start2 = start1.astimezone(tz)

        r1 = rrule('MINUTELY', dtstart=start1, count=1)
        r2 = rrule('MINUTELY', dtstart=start2, count=1)

        r1_json = self.dumps(r1)
        r2_json = self.dumps(r2)

        r1_parsed = self.loads(r1_json)
        self.assertEqual(r1_parsed.dtstart.utcoffset(), r1.dtstart.utcoffset())

        r2_parsed = self.loads(r2_json)
        self.assertEqual(r2_parsed.dtstart.utcoffset(), r2.dtstart.utcoffset())

        self.assertEqual(r1_parsed.dtstart, r2_parsed.dtstart)
        self.assertNotEqual(r1_parsed.dtstart.utcoffset(), r2_parsed.dtstart.utcoffset())

    def test_weekday(self):
        w = dateutil_rrule.weekday(5)
        result = self.dumps(w)
        self.assertEqual(json.loads(result), self.weekday())

    def test_weekday_in_rrule(self):
        r = rrule(
            dateutil_rrule.WEEKLY,
            dtstart=datetime(2015, 12, 30, 12, 59, 22, tzinfo=timezone.utc),
            until=datetime(2015, 12, 31, 12, 59, 22, tzinfo=timezone.utc),
            byweekday=(dateutil_rrule.MO, dateutil_rrule.WE, dateutil_rrule.FR),
        )
        result = self.dumps(r)
        self.assertEqual(
            json.loads(result),
            {
                '__type__': 'rrule',
                'freq': 2,
                'dtstart': 1451480362,
                'dtstart_tz': 0,
                'interval': 1,
                'wkst': None,
                'count': None,
                'until': 1451566762,
                'until_tz': 0,
                'bysetpos': None,
                'bymonth': None,
                'bymonthday': None,
                'byyearday': None,
                'byeaster': None,
                'byweekno': None,
                'byweekday': [
                    {'__type__': 'weekday', 'wkday': 0},
                    {'__type__': 'weekday', 'wkday': 2},
                    {'__type__': 'weekday', 'wkday': 4},
                ],
                'byhour': None,
                'byminute': None,
                'bysecond': None,
            },
        )


class RedBeatJSONDecoderTestCase(JSONTestCase):
    def test_datetime_no_timezone(self):
        d = self.datetime_as_dict(__type__='datetime')
        result = self.loads(json.dumps(d))
        d.pop('__type__')
        self.assertEqual(result, datetime(tzinfo=timezone.utc, **d))

    def test_datetime_with_timezone(self):
        d = self.datetime_as_dict(__type__='datetime', timezone='Asia/Shanghai')
        result = self.loads(json.dumps(d))
        d.pop('__type__')
        d.pop('timezone')
        self.assertEqual(result, datetime(tzinfo=zoneinfo.ZoneInfo('Asia/Shanghai'), **d))

    def test_datetime_with_fixed_offset(self):
        d = self.datetime_as_dict(__type__='datetime', timezone=5 * 60 * 60)
        result = self.loads(json.dumps(d))
        d.pop('__type__')
        d.pop('timezone')
        self.assertEqual(result, datetime(tzinfo=timezone(timedelta(hours=5)), **d))

    def test_schedule(self):
        d = self.schedule()

        result = self.loads(json.dumps(d))
        self.assertEqual(result, schedule(run_every=60))

    def test_crontab(self):
        d = self.crontab()

        result = self.loads(json.dumps(d))
        self.assertEqual(result, crontab())

    def test_rrule(self):
        d = self.rrule()

        result = self.loads(json.dumps(d))
        self.assertEqual(
            result,
            rrule(
                'MINUTELY', dtstart=datetime(2015, 12, 30, 12, 59, 22, tzinfo=timezone.utc), count=1
            ),
        )

    def test_weekday(self):
        d = self.weekday()

        result = self.loads(json.dumps(d))
        self.assertEqual(result, dateutil_rrule.weekday(5))



BROKER_URL = 'redis://'
# ELERY_ENABLE_UTC = True
# CELERY_TIMEZONE = 'UTC'

CELERYBEAT_SCHEDULER = 'redbeat.RedBeatScheduler'
CELERYBEAT_MAX_LOOP_INTERVAL = 5  # redbeat likes fast loops

# REDBEAT_REDIS_URL = BROKER_URL
# REDBEAT_KEY_PREFIX = 'redbeat'
# REDBEAT_SCHEDULE_KEY = REDBEAT_KEY_PREFIX + ':schedule'
# REDBEAT_STATICS_KEY = REDBEAT_KEY_PREFIX + ':statics'
# REDBEAT_LOCK_KEY = REDBEAT_KEY_PREFIX + ':lock'
# REDBEAT_LOCK_TIMEOUT = CELERYBEAT_MAX_LOOP_INTERVAL * 5
